/*
 * Copyright 2002 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * This file is part of the FSG PAPI Project
 *
 * The FSG PAPI Project is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The FSG PAPI Project is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with the FSG PAPI Project; see the file COPYING.LIB.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* $Id: attribute.c,v 1.3 2002/10/24 06:13:23 jacobs Exp $ */
#pragma ident	"@(#)attribute.c	1.1	02/08/22 SMI"

/*LINTLIBRARY*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <papi.h>

static void papiAttributeFree(papi_attribute_t *attribute);

static void
papiAttributeValueFree(papi_attribute_value_type_t type,
		    papi_attribute_value_t *value)
{
	if (value != NULL) {
		switch (type) {
		case PAPI_STRING:
			if (value->string != NULL)
				free(value->string);
			break;
		case PAPI_COLLECTION:
			if (value->collection != NULL) {
				int i;

				for (i = 0; value->collection[i] != NULL; i++)
					papiAttributeFree(value->collection[i]);
	
				free(value->collection);
			}
			break;
		default: /* don't need to free anything extra */
			break;
		}

		free(value);
	}
}

static void
papiAttributeValuesFree(papi_attribute_value_type_t type,
		    papi_attribute_value_t **values)
{
	if (values != NULL) {
		int i;

		for (i = 0 ; values[i] != NULL ; i++)
			papiAttributeValueFree(type, values[i]);

		free(values);
	}
}

static void
papiAttributeFree(papi_attribute_t *attribute)
{
	if (attribute != NULL) {
		if (attribute->name != NULL)
			free(attribute->name);
		if (attribute->values != NULL)
			papiAttributeValuesFree(attribute->type,
						attribute->values);
			free(attribute);
	}
}

void
papiAttributeListFree(papi_attribute_t **list)
{
	if (list != NULL) {
		int i;

		for (i = 0 ; list[i] != NULL ; i++)
			papiAttributeFree(list[i]);

		free(list);
	}
}

static papi_attribute_t **
collection_dup(papi_attribute_t **collection)
{
	papi_attribute_t **result = NULL;

	if (collection != NULL) {
		int i;

		for (i = 0; collection[i] != NULL; i++) {
			papi_attribute_t *a = collection[i];

			papiAttributeListAdd(&result, PAPI_ATTR_APPEND, a->name,
					a->type, NULL);
			if (a->values != NULL) {
				int j;

				for (j = 0; a->values[j] != NULL; j++)
					papiAttributeListAdd(&result,
							PAPI_ATTR_APPEND,
							a->name, a->type,
							a->values[j]);
			}
		}
	}

	return (result);
}

static papi_attribute_value_t *
papiAttributeValueDup(const papi_attribute_value_type_t type,
		const papi_attribute_value_t *v)
{
	papi_attribute_value_t *result = NULL;

	if ((v != NULL) && ((result = calloc(1, sizeof (*result))) != NULL)) {
		switch (type) {
		case PAPI_STRING:
			if (v->string == NULL) {
				free(result);
				result = NULL;
			} else
				result->string = strdup(v->string);
			break;
		case PAPI_INTEGER:
			result->integer = v->integer;
			break;
		case PAPI_BOOLEAN:
			result->boolean = v->boolean;
			break;
		case PAPI_RANGE:
			result->range.lower = v->range.lower;
			result->range.upper = v->range.upper;
			break;
		case PAPI_RESOLUTION:
			result->resolution.xres = v->resolution.xres;
			result->resolution.yres = v->resolution.yres;
			result->resolution.units = v->resolution.units;
			break;
		case PAPI_DATETIME:
			result->datetime = v->datetime;
			break;
		case PAPI_COLLECTION:
			result->collection = collection_dup(v->collection);
			break;
		default:	/* unknown type, fail to duplicate */
			free(result);
			result = NULL;
		}
	}

	return (result);
}

static papi_attribute_t *
papiAttributeAlloc(const char *name, papi_attribute_value_type_t type)
{
	papi_attribute_t *result = NULL;

	if ((result = calloc(1, sizeof (*result))) != NULL) {
		result->name = strdup(name);
		result->type = type;
	}

	return (result);
}

static papi_status_t
papiAttributeListAddValue(papi_attribute_value_t ***values,
		      const papi_attribute_value_type_t type,
		      const papi_attribute_value_t *value)
{

	if (values == NULL)
		return (PAPI_BAD_ARGUMENT);

	if (value != NULL) {	/* this allows "empty" attributes */
		papi_attribute_value_t *tmp = NULL;

		if ((tmp = papiAttributeValueDup(type, value)) == NULL)
			return (PAPI_TEMPORARY_ERROR);

		list_append(values, tmp);
	}

	return (PAPI_OK);
}

papi_status_t
papiAttributeListAdd(papi_attribute_t ***list, const int flgs,
		const char *name, const papi_attribute_value_type_t type,
		const papi_attribute_value_t *value)
{
	papi_status_t result;
	int flags = flgs;
	papi_attribute_t *attribute = NULL;

	if ((list == NULL) || (name == NULL))
		return (PAPI_BAD_ARGUMENT);

	if ((type == PAPI_RANGE) && (value != NULL) &&
	    (value->range.lower > value->range.upper))
		return (PAPI_BAD_ARGUMENT);	/* RANGE must have min <= max */

	if (flags == 0) /* if it wasn't set, set a default behaviour */
		flags = PAPI_ATTR_APPEND;

	/* look for an existing one */
	attribute = papiAttributeListFind(*list, name);

	if (((flags & PAPI_ATTR_EXCL) != 0) && (attribute != NULL))
		return (PAPI_CONFLICT); /* EXISTS */

	if (((flags & PAPI_ATTR_REPLACE) == 0) && (attribute != NULL) &&
	    (attribute->type != type))
		return (PAPI_CONFLICT); /* TYPE CONFLICT */

	/* if we don't have one, create it and add it to the list */
	if ((attribute == NULL) &&
	    ((attribute = papiAttributeAlloc(name, type)) != NULL))
		list_append(list, attribute);

	/* if we don't have one by now, it's most likely an alloc fail */
	if (attribute == NULL)
		return (PAPI_TEMPORARY_ERROR);

	/* if we are replacing, clear any existing values */
	if (((flags & PAPI_ATTR_REPLACE) != 0) && (attribute->values != NULL)) {
		papiAttributeValuesFree(attribute->type, attribute->values);
		attribute->values = NULL;
	}

	attribute->type = type;

	result = papiAttributeListAddValue(&attribute->values, type, value);

	return (result);
}

papi_status_t
papiAttributeListAddString(papi_attribute_t ***list, const int flags,
			const char *name, const char *string)
{
	papi_attribute_value_t v;

	v.string = (char *)string;
	return (papiAttributeListAdd(list, flags, name, PAPI_STRING, &v));
}

papi_status_t
papiAttributeListAddInteger(papi_attribute_t ***list, const int flags,
			const char *name, const int integer)
{
	papi_attribute_value_t v;

	v.integer = integer;
	return (papiAttributeListAdd(list, flags, name, PAPI_INTEGER, &v));
}

papi_status_t
papiAttributeListAddBoolean(papi_attribute_t ***list, const int flags,
			const char *name, const char boolean)
{
	papi_attribute_value_t v;

	v.boolean = boolean;
	return (papiAttributeListAdd(list, flags, name, PAPI_BOOLEAN, &v));
}

papi_status_t
papiAttributeListAddRange(papi_attribute_t ***list, const int flags,
			const char *name, const int lower, const int upper)
{
	papi_attribute_value_t v;

	v.range.lower = lower;
	v.range.upper = upper;
	return (papiAttributeListAdd(list, flags, name, PAPI_RANGE, &v)); 
}

papi_status_t
papiAttributeListAddResolution(papi_attribute_t ***list, const int flags,
			const char *name, const int xres, const int yres,
			const papi_resolution_unit_t units)
{
	papi_attribute_value_t v;

	v.resolution.xres = xres;
	v.resolution.yres = yres;
	v.resolution.units = units;
	return (papiAttributeListAdd(list, flags, name, PAPI_RESOLUTION, &v));
}

papi_status_t
papiAttributeListAddDatetime(papi_attribute_t ***list, const int flags,
			const char *name, const time_t datetime)
{
	papi_attribute_value_t v;

	v.datetime = datetime;
	return (papiAttributeListAdd(list, flags, name, PAPI_DATETIME, &v));
}

papi_status_t
papiAttributeListAddCollection(papi_attribute_t ***list, const int flags,
			const char *name, const papi_attribute_t **collection)
{
	papi_attribute_value_t v;

	v.collection = (papi_attribute_t **)collection;
	return (papiAttributeListAdd(list, flags, name, PAPI_COLLECTION, &v));
}

papi_status_t
papiAttributeDelete(papi_attribute_t ***list, const char *name)
{
	papi_attribute_t *attribute;

	if ((list == NULL) || (name == NULL))
		return (PAPI_BAD_ARGUMENT);

	if ((attribute = papiAttributeListFind(*list, name)) == NULL)
		return (PAPI_NOT_FOUND);

	list_remove(*list, attribute);
	papiAttributeFree(attribute);

	return (PAPI_OK);
}

papi_attribute_t *
papiAttributeListFind(papi_attribute_t **list, const char *name)
{
	if (list != NULL) {
		int i;

		for (i = 0 ; list[i] != NULL ; i++)
			if (strcasecmp(list[i]->name, name) == 0)
				return ((papi_attribute_t *)list[i]);
	}

	return (NULL);
}

papi_attribute_t *
papiAttributeListGetNext(papi_attribute_t **list, void **iter)
{
	papi_attribute_t **tmp, *result;

	if ((list == NULL) && (iter == NULL))
		return (NULL);

	if (*iter == NULL)
		*iter = list;

	tmp = *iter;
	result = *tmp;
	*iter = ++tmp;

	return (result);
}

papi_status_t
papiAttributeListGetValue(papi_attribute_t **list, void **iter,
			const char *name, papi_attribute_value_type_t type,
			papi_attribute_value_t **value)
{
	papi_attribute_value_t **tmp;
	void *fodder = NULL;

	if ((list == NULL) || ((name == NULL) && (iter == NULL)) ||
	    (value == NULL))
		return (PAPI_BAD_ARGUMENT);

	if (iter == NULL)
		iter = &fodder;

	if (name != NULL) {
		papi_attribute_t *attr = papiAttributeListFind(list, name);

		if (attr == NULL)
			return (PAPI_NOT_FOUND);

		if (attr->type != type)
			return (PAPI_NOT_POSSIBLE);

		tmp = attr->values;
	} else
		tmp = *iter;

	if (tmp == NULL)
		return (PAPI_NOT_FOUND);

	*value = *tmp;
	*iter =  ++tmp;

	if (*value == NULL)
		return (PAPI_GONE);

	return (PAPI_OK);
}

papi_status_t
papiAttributeListGetString(papi_attribute_t **list, void **iter,
			const char *name, char **vptr)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if (vptr == NULL)
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_STRING, &value);
	if (status == PAPI_OK)
		*vptr = value->string;

	return (status);
}

papi_status_t
papiAttributeListGetInteger(papi_attribute_t **list, void **iter,
			const char *name, int *vptr)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if (vptr == NULL)
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_INTEGER, &value);
	if (status == PAPI_OK)
		*vptr = value->integer;

	return (status);
}

papi_status_t
papiAttributeListGetBoolean(papi_attribute_t **list, void **iter,
			const char *name, char *vptr)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if (vptr == NULL)
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_BOOLEAN, &value);
	if (status == PAPI_OK)
		*vptr = value->boolean;

	return (status);
}

papi_status_t
papiAttributeListGetRange(papi_attribute_t **list, void **iter,
			const char *name, int *min, int *max)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if ((min == NULL) || (max == NULL))
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_RANGE, &value);
	if (status == PAPI_OK) {
		*min= value->range.lower;
		*max = value->range.upper;
	}

	return (status);
}

papi_status_t
papiAttributeListGetResolution(papi_attribute_t **list, void **iter,
			const char *name, int *x, int *y,
			papi_resolution_unit_t *units)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if ((x == NULL) || (y == NULL) || (units == NULL))
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_RESOLUTION, &value);
	if (status == PAPI_OK) {
		*x= value->resolution.xres;
		*y = value->resolution.yres;
		*units = value->resolution.units;
	}

	return (status);
}

papi_status_t
papiAttributeListGetDatetime(papi_attribute_t **list, void **iter,
			const char *name, time_t *dt)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if (dt == NULL)
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_DATETIME, &value);
	if (status == PAPI_OK) {
		*dt= value->datetime;
	}

	return (status);
}

papi_status_t
papiAttributeListGetCollection(papi_attribute_t **list, void **iter,
			const char *name, papi_attribute_t ***collection)
{
	papi_status_t status;
	papi_attribute_value_t *value = NULL;

	if (collection == NULL)
		return (PAPI_BAD_ARGUMENT);

	status = papiAttributeListGetValue(list, iter, name,
				PAPI_COLLECTION, &value);
	if (status == PAPI_OK) {
		*collection= value->collection;
	}

	return (status);
}

static char *
typeString(papi_attribute_value_type_t type)
{
	switch (type) {
	case PAPI_STRING:
		return ("string");
	case PAPI_INTEGER:
		return ("integer");
	case PAPI_BOOLEAN:
		return ("boolean");
	case PAPI_RANGE:
		return ("range");
	case PAPI_RESOLUTION:
		return ("resolution");
	case PAPI_DATETIME:
		return ("datetime");
	case PAPI_COLLECTION:
		return ("collection");
	}

	return ("unknown");


}

papi_status_t papiAttributeListFromString(papi_attribute_t ***attrs,
		const int flags, const char *string)
{
	return (PAPI_OPERATION_NOT_SUPPORTED);
}

extern papi_status_t papiAttributeListToString(const papi_attribute_t **attrs,
		const char *delim,
		char *buffer, const size_t buflen)
{
	return (PAPI_OPERATION_NOT_SUPPORTED);
}

/* for debugging, not part of the public API */
void papiAttributeListPrint(FILE *fp, char *prefix, papi_attribute_t **list);

void
papiAttributePrint(FILE *fp, char *prefix, papi_attribute_t *attribute)
{
	if (attribute != NULL) {
		char *name = attribute->name;

		if (prefix == NULL)
			prefix = "";

		fprintf(fp, "%s'%s' (%s) =\n", prefix, (name ? name : "(NULL)"),
			typeString(attribute->type));
		if (attribute->values != NULL) {
			papi_attribute_value_t **values = attribute->values;
			int i;

			for (i = 0; values[i] != NULL; i++) {
				if (attribute->type != PAPI_COLLECTION)
					fprintf(fp, "%s\t", prefix);
				switch (attribute->type) {
				case PAPI_STRING:
					fprintf(fp, "%s", values[i]->string);
					break;
				case PAPI_INTEGER:
					fprintf(fp, "%d", values[i]->integer);
					break;
				case PAPI_BOOLEAN:
					fprintf(fp, "%s", (values[i]->boolean ?
							   "true" : "false"));
					break;
				case PAPI_RANGE:
					fprintf(fp, "(%d - %d)",
						values[i]->range.lower,
						values[i]->range.upper);
					break;
				case PAPI_RESOLUTION:
					fprintf(fp, "%dx%d",
						values[i]->resolution.xres,
						values[i]->resolution.yres);
					
					break;
				case PAPI_DATETIME: {
					struct tm *tm;

					tm = localtime(&values[i]->datetime);
					if (tm != NULL) {
						char buf[64];

                				strftime(buf, sizeof (buf),
							"%C", tm);
						fprintf(fp, "%s", buf);
					}}
					break;
				case PAPI_COLLECTION: {
					char s[64];

					snprintf(s, sizeof (s), "%s\t", prefix);
					papiAttributeListPrint(fp, s,
							values[i]->collection);
					}
					break;
				default:
					fprintf(fp, "unknown");
				}
				if (attribute->type != PAPI_COLLECTION)
					fprintf(fp, "\n");
			}
		}
	}
}


void
papiAttributeListPrint(FILE *fp, char *prefix, papi_attribute_t **list)
{
	if (list != NULL) {
		int i;

		for (i = 0; list[i] != NULL; i++)
			papiAttributePrint(fp, prefix, list[i]);
	} else
		fprintf(fp, "empty list\n");
}
